@@ -66,36 +66,42 @@ static struct pam_conv conv = { NULL, NULL};
 
 const char *applet = NULL;
 const char *extraopts = NULL;
-const char *getoptstring = "d:e:g:I:Kk:N:p:r:Su:1:2:" \
+const char *getoptstring = "D:d:e:g:I:Kk:m:N:p:r:Su:1:2:" \
 	getoptstring_COMMON;
 const struct option longopts[] = {
+	{ "respawn-delay",        1, NULL, 'D'},
 	{ "chdir",        1, NULL, 'd'},
 	{ "env",          1, NULL, 'e'},
 	{ "group",        1, NULL, 'g'},
 	{ "ionice",       1, NULL, 'I'},
 	{ "stop",         0, NULL, 'K'},
 	{ "umask",        1, NULL, 'k'},
+	{ "respawn-max",    1, NULL, 'm'},
 	{ "nicelevel",    1, NULL, 'N'},
 	{ "pidfile",      1, NULL, 'p'},
-	{ "user",         1, NULL, 'u'},
+	{ "respawn-period",        1, NULL, 'P'},
 	{ "chroot",       1, NULL, 'r'},
 	{ "start",        0, NULL, 'S'},
+	{ "user",         1, NULL, 'u'},
 	{ "stdout",       1, NULL, '1'},
 	{ "stderr",       1, NULL, '2'},
 	longopts_COMMON
 };
 const char * const longopts_help[] = {
+	"Set a respawn delay",
 	"Change the PWD",
 	"Set an environment string",
 	"Change the process group",
 	"Set an ionice class:data when starting",
 	"Stop daemon",
 	"Set the umask for the daemon",
+	"set maximum number of respawn attempts",
 	"Set a nicelevel when starting",
 	"Match pid found in this file",
-	"Change the process user",
+	"Set respawn time period",
 	"Chroot to this directory",
 	"Start daemon",
+	"Change the process user",
 	"Redirect stdout to file",
 	"Redirect stderr to file",
 	longopts_help_COMMON
@@ -165,7 +171,8 @@ static pid_t get_pid(const char *pidfile)
 	return pid;
 }
 
-static void child_process(char *exec, char **argv)
+static void child_process(char *exec, char **argv, char *svcname,
+		int start_count)
 {
 	RC_STRINGLIST *env_list;
 	RC_STRING *env;
@@ -177,6 +184,9 @@ static void child_process(char *exec, char **argv)
 	char *np;
 	char **c;
 	char cmdline[PATH_MAX];
+	time_t start_time;
+	char start_count_string[20];
+	char start_time_string[20];
 
 #ifdef HAVE_PAM
 	pam_handle_t *pamh = NULL;
@@ -186,6 +196,14 @@ static void child_process(char *exec, char **argv)
 
 	setsid();
 
+	if (svcname) {
+start_time = time(NULL);
+from_time_t(start_time_string, start_time);
+		rc_service_value_set(svcname, "start_time", start_time_string);
+sprintf(start_count_string, "%i", start_count);
+		rc_service_value_set(svcname, "start_count", start_count_string);
+	}
+
 	if (nicelevel) {
 		if (setpriority(PRIO_PROCESS, getpid(), nicelevel) == -1)
 			eerrorx("%s: setpriority %d: %s", applet, nicelevel,
@@ -226,6 +244,7 @@ static void child_process(char *exec, char **argv)
 	/* Close any fd's to the passwd database */
 	endpwent();
 
+	/* remove the controlling tty */
 #ifdef TIOCNOTTY
 	ioctl(tty_fd, TIOCNOTTY, 0);
 	close(tty_fd);
@@ -321,7 +340,7 @@ static void child_process(char *exec, char **argv)
 		dup2(stderr_fd, STDERR_FILENO);
 
 	for (i = getdtablesize() - 1; i >= 3; --i)
-		close(i);
+		fcntl(i, F_SETFD, FD_CLOEXEC);
 
 	*cmdline = '\0';
 	c = argv;
@@ -423,7 +442,14 @@ int main(int argc, char **argv)
 	char *p;
 	char *token;
 	int i;
+	int n;
 	char exec_file[PATH_MAX];
+	int respawn_count = 0;
+	int respawn_delay = 0;
+	int respawn_max = 10;
+	int respawn_period = 5;
+	time_t respawn_now= 0;
+	time_t first_spawn= 0;
 	struct passwd *pw;
 	struct group *gr;
 	FILE *fp;
@@ -460,6 +486,12 @@ int main(int argc, char **argv)
 	while ((opt = getopt_long(argc, argv, getoptstring, longopts,
 		    (int *) 0)) != -1)
 		switch (opt) {
+		case 'D':  /* --respawn-delay time */
+			n = sscanf(optarg, "%d", &respawn_delay);
+			if (n	!= 1 || respawn_delay < 1)
+				eerrorx("Invalid respawn-delay value '%s'", optarg);
+			break;
+
 		case 'I': /* --ionice */
 			if (sscanf(optarg, "%d:%d", &ionicec, &ioniced) == 0)
 				eerrorx("%s: invalid ionice `%s'",
@@ -481,6 +513,12 @@ int main(int argc, char **argv)
 				    applet, optarg);
 			break;
 
+		case 'P':  /* --respawn-period time */
+			n = sscanf(optarg, "%d", &respawn_period);
+			if (n	!= 1 || respawn_delay < 1)
+				eerrorx("Invalid respawn-delay value '%s'", optarg);
+			break;
+
 		case 'S':  /* --start */
 			start = true;
 			break;
@@ -510,6 +548,12 @@ int main(int argc, char **argv)
 				    applet, optarg);
 			break;
 
+		case 'm':  /* --respawn-max count */
+			n = sscanf(optarg, "%d", &respawn_max);
+			if (n	!= 1 || respawn_max < 1)
+				eerrorx("Invalid respawn-max value '%s'", optarg);
+			break;
+
 		case 'p':  /* --pidfile <pid-file> */
 			pidfile = optarg;
 			break;
@@ -580,6 +624,11 @@ int main(int argc, char **argv)
 	if (start) {
 		if (!exec)
 			eerrorx("%s: nothing to start", applet);
+		if (respawn_delay * respawn_max > respawn_period) {
+			ewarn("%s: Please increase the value of --respawn-period to more "
+				"than %d to avoid infinite respawning", applet, 
+				respawn_delay * respawn_max);
+		}
 	}
 
 	/* Expand ~ */
@@ -639,10 +688,12 @@ int main(int argc, char **argv)
 		 * result would be the same. */
 		if (pidfile && exists(pidfile))
 			unlink(pidfile);
-		if (svcname)
+		if (svcname) {
 			rc_service_daemon_set(svcname, exec,
 			    (const char *const *)argv,
 			    pidfile, false);
+			rc_service_mark(svcname, RC_SERVICE_STOPPED);
+		}
 		exit(EXIT_SUCCESS);
 	}
 
@@ -674,6 +725,10 @@ int main(int argc, char **argv)
 	if (child_pid != 0)
 		exit(EXIT_SUCCESS);
 
+#ifdef TIOCNOTTY
+	tty_fd = open("/dev/tty", O_RDWR);
+#endif
+	devnull_fd = open("/dev/null", O_RDWR);
 	child_pid = fork();
 	if (child_pid == -1)
 		eerrorx("%s: fork: %s", applet, strerror(errno));
@@ -682,18 +737,22 @@ int main(int argc, char **argv)
 		/* this is the supervisor */
 		umask(numask);
 
-#ifdef TIOCNOTTY
-		tty_fd = open("/dev/tty", O_RDWR);
-#endif
-
-		devnull_fd = open("/dev/null", O_RDWR);
-
 		fp = fopen(pidfile, "w");
 		if (! fp)
 			eerrorx("%s: fopen `%s': %s", applet, pidfile, strerror(errno));
 		fprintf(fp, "%d\n", getpid());
 		fclose(fp);
 
+		if (svcname)
+			rc_service_daemon_set(svcname, exec,
+									(const char * const *) argv, pidfile, true);
+
+	/* remove the controlling tty */
+#ifdef TIOCNOTTY
+		ioctl(tty_fd, TIOCNOTTY, 0);
+		close(tty_fd);
+#endif
+
 		/*
 		 * Supervisor main loop
 		 */
@@ -704,6 +763,23 @@ int main(int argc, char **argv)
 				syslog(LOG_INFO, "stopping %s, pid %d", exec, child_pid);
 				kill(child_pid, SIGTERM);
 			} else {
+				sleep(respawn_delay);
+				if (respawn_max > 0 && respawn_period > 0) {
+					respawn_now = time(NULL);
+					if (first_spawn == 0)
+						first_spawn = respawn_now;
+					if (respawn_now - first_spawn > respawn_period) {
+						respawn_count = 0;
+						first_spawn = 0;
+					} else
+						respawn_count++;
+					if (respawn_count >= respawn_max) {
+						syslog(LOG_INFO, "respawned \"%s\" too many times, "
+								"exiting", exec);
+						exiting = true;
+						continue;
+					}
+				}
 				if (WIFEXITED(i))
 					syslog(LOG_INFO, "%s, pid %d, exited with return code %d",
 							exec, child_pid, WEXITSTATUS(i));
@@ -714,15 +790,19 @@ int main(int argc, char **argv)
 				if (child_pid == -1)
 					eerrorx("%s: fork: %s", applet, strerror(errno));
 				if (child_pid == 0)
-					child_process(exec, argv);
+					child_process(exec, argv, svcname, respawn_count);
 			}
 		}
 
-		if (svcname)
+		if (pidfile && exists(pidfile))
+			unlink(pidfile);
+		if (svcname) {
 			rc_service_daemon_set(svcname, exec,
-									(const char * const *) argv, pidfile, true);
-
+			    (const char *const *)argv,
+			    pidfile, false);
+			rc_service_mark(svcname, RC_SERVICE_STOPPED);
+		}
 		exit(EXIT_SUCCESS);
 	} else if (child_pid == 0)
-		child_process(exec, argv);
+		child_process(exec, argv, svcname, respawn_count);
 }
